/* -*- Mode: C++; tab-width: 2; indent-tabs-mode: nil; c-basic-offset: 2 -*- */
/* ***** BEGIN LICENSE BLOCK *****
 * Version: MPL 1.1/GPL 2.0/LGPL 2.1
 *
 * The contents of this file are subject to the Mozilla Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is the mozCellML code.
 *
 * The Initial Developer of the Original Code is
 * The Bioengineering Institute, University of Auckland.
 * Portions created by the Initial Developer are Copyright (C) 2004
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):
 *   Andrew Miller <ak.miller@auckland.ac.nz> (original developer)
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either the GNU General Public License Version 2 or later (the "GPL"), or
 * the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the MPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the MPL, the GPL or the LGPL.
 *
 * ***** END LICENSE BLOCK ***** */

#include "mozCellML.h"
#include <nsIDOMElement.h>
#include <nsCOMPtr.h>
#include <nsString.h>
#include <nsCOMArray.h>
#include <nsPrintfCString.h>
#include <nsIArray.h>
#include <nsDataHashtable.h>
#include <nsInterfaceHashtable.h>
#include <math.h>
#include "mozCellMLUnits.h"

NS_IMPL_ISUPPORTS0(mozCellMLUnits)
NS_IMPL_ISUPPORTS0(mozCellMLUnits::UnitBase)

#define T(model, s) (model->Translate(s))

#define COMPLAIN(model,args) \
{\
  char* result = PR_smprintf args; \
  nsCAutoString complainCString(result); \
  nsAutoString complainString; \
  CopyUTF8toUTF16(complainCString, complainString); \
  PR_Free(result); \
  return model->ValidationError(complainString); \
}

mozCellMLUnits::mozCellMLUnits()
  : isbase(PR_TRUE), seen(PR_TRUE), done(PR_TRUE),
    multiplier(1.0), offset(0.0)
{
}

mozCellMLUnits::mozCellMLUnits(mozCellMLUnits* u,
                               double exp,
                               double ioffset,
                               double imup)
  : isbase(PR_FALSE), seen(PR_TRUE), done(PR_TRUE),
    multiplier(imup), offset(ioffset)
{
  nsCOMPtr<UnitBase> ub = new UnitBase(u, exp);
  unitList.AppendObject(ub);
}

mozCellMLUnits::mozCellMLUnits(
                               mozCellMLUnits* u1,
                               double exp1,
                               mozCellMLUnits* u2,
                               double exp2
                              )
  : isbase(PR_FALSE), seen(PR_TRUE), done(PR_TRUE),
    multiplier(1.0), offset(0.0)
{
  if (u1 > u2)
  {
    mozCellMLUnits* utmp = u1;
    double etmp = exp1;
    u1 = u2;
    exp1 = exp2;
    u2 = utmp;
    exp2 = etmp;
  }
  nsCOMPtr<UnitBase> ub1 = new UnitBase(u1, exp1);
  nsCOMPtr<UnitBase> ub2 = new UnitBase(u2, exp2);
  unitList.AppendObject(ub1);
  unitList.AppendObject(ub2);
}

mozCellMLUnits::mozCellMLUnits(
                               mozCellMLUnits* u1,
                               double exp1,
                               mozCellMLUnits* u2,
                               double exp2,
                               mozCellMLUnits* u3,
                               double exp3
                              )
  : isbase(PR_FALSE), seen(PR_TRUE), done(PR_TRUE),
    multiplier(1.0), offset(0.0)
{
  double xexp1 = 0 , xexp2 = 0, xexp3 = 0;
  mozCellMLUnits* xu1 = nsnull, *xu2 = nsnull, *xu3 = nsnull;
  
#define ORDER(i,j,k) \
  xexp1 = exp##i; \
  xexp2 = exp##j; \
  xexp3 = exp##k; \
  xu1 = u##i; \
  xu2 = u##j; \
  xu3 = u##k;

  switch (((u1 > u2) ? 4 : 0) |
          ((u1 > u3) ? 2 : 0) |
          ((u2 > u3) ? 1 : 0))
  {
  case 0:
    ORDER(1, 2, 3);
    break;
  case 1:
    ORDER(1, 3, 2);
    break;
  case 3:
    ORDER(3, 1, 2);
    break;
  case 4:
    ORDER(2, 1, 3);
    break;
  case 6:
    ORDER(2, 3, 1);
    break;
  case 7:
    ORDER(3, 2, 1);
    break;
  }
#undef ORDER

  nsCOMPtr<UnitBase> ub1 = new UnitBase(xu1, xexp1);
  nsCOMPtr<UnitBase> ub2 = new UnitBase(xu2, xexp2);
  nsCOMPtr<UnitBase> ub3 = new UnitBase(xu3, xexp3);
  unitList.AppendObject(ub1);
  unitList.AppendObject(ub2);
  unitList.AppendObject(ub3);
}

mozCellMLUnits::mozCellMLUnits(
                               mozCellMLUnits* u1,
                               double exp1,
                               mozCellMLUnits* u2,
                               double exp2,
                               mozCellMLUnits* u3,
                               double exp3,
                               mozCellMLUnits* u4,
                               double exp4
                              )
  : isbase(PR_FALSE), seen(PR_TRUE), done(PR_TRUE),
    multiplier(1.0), offset(0.0)
{
  nsCOMPtr<UnitBase> ub1 = new UnitBase(u1, exp1);
  nsCOMPtr<UnitBase> ub2 = new UnitBase(u2, exp2);
  nsCOMPtr<UnitBase> ub3 = new UnitBase(u3, exp3);
  nsCOMPtr<UnitBase> ub4 = new UnitBase(u4, exp4);
  unitList.AppendObject(ub1);
  unitList.AppendObject(ub2);
  unitList.AppendObject(ub3);
  unitList.AppendObject(ub4);
  SortBases(0, 3);
}

mozCellMLUnits::mozCellMLUnits(
                               PRUint32 imp,
                               const nsAString& comp,
                               nsIDOMElement* itag
                              )
  : isbase(PR_FALSE), seen(PR_FALSE), done(PR_FALSE), importNo(imp),
    component(comp), tag(itag)
{
  tag->GetAttribute(NS_LITERAL_STRING("name"), name);

  nsAutoString baseUnits;
  tag->GetAttribute(NS_LITERAL_STRING("base_units"), baseUnits);
  if (baseUnits.Equals(NS_LITERAL_STRING("yes")))
  {
    isbase = PR_TRUE;
    done = PR_TRUE;
    seen = PR_TRUE;
    multiplier = 1.0;
    offset = 0.0;
    return;
  }
}

void
mozCellMLUnits::GetFQName(nsAString& fqname)
{
  nsPrintfCString inCS("%u:", importNo);
  CopyUTF8toUTF16(inCS, fqname);
  fqname.Append(component);
  fqname.Append(NS_LITERAL_STRING(":"));
  fqname.Append(name);
}

// True if b1 base > b2 base...
PRBool
mozCellMLUnits::CompareBases(UnitBase* b1, UnitBase* b2)
{
  return (((void*)b1->base) > ((void*)b2->base));
}

void
mozCellMLUnits::SwapBases(PRUint32 idx1, PRUint32 idx2)
{
  nsCOMPtr<UnitBase> tmp = unitList[idx1];
  unitList.ReplaceObjectAt(unitList[idx2], idx1);
  unitList.ReplaceObjectAt(tmp, idx2);
}

// Put the base units in order(of pointer). This makes it O(n) in the number of
// base units to subsequently compare this unit with others.
void
mozCellMLUnits::SortBases(PRUint32 first, PRUint32 last)
{
  if (first >= last)
  {
    return;
  }
  if ((last - first) == 1)
  {
    if (CompareBases(unitList[first], unitList[last]))
      SwapBases(first, last);
    return;
  }
  PRUint32 mid = (last + first) >> 1;
  UnitBase* midBase = unitList[mid];
  SwapBases(last, mid);

  mid = first;
  PRUint32 i;
  for (i = first; i < last; i++)
    if (CompareBases(midBase, unitList[i]))
      SwapBases(i, mid++);
  SwapBases(mid, last);
  if (mid > 0)
    SortBases(first, mid - 1);
  SortBases(mid + 1, last);
}

nsresult
mozCellMLUnits::Initialise(
                           mozCellML* model,
                           nsDataHashtable<nsISupportsHashKey,
                                           PRUint32>&
                             importNumbers,
                           nsInterfaceHashtable<nsStringHashKey,
                                                mozCellMLUnits>&
                             allUnits
                          )
{
  seen = PR_TRUE;
  if (isbase)
  {
    done = PR_TRUE;
    return NS_OK;
  }
  if (name.Equals(NS_LITERAL_STRING("")))
    COMPLAIN(model, (T(model, "nameAttributeMandatory")));
  
  // Next, find all our dependant units, and initialise them if needed...
  nsCOMPtr<nsIDOMNodeList> nl;
  tag->GetChildNodes(getter_AddRefs(nl));
  nsCOMPtr<nsIArray> arr;
  FilterNodeListByNameAndTypeNS2(nl, getter_AddRefs(arr),
                                 nsIDOMNode::ELEMENT_NODE,
                                 CELLML_NS_1_0,
                                 CELLML_NS_1_1,
                                 NS_LITERAL_STRING("unit"));
  nsCOMPtr<nsISimpleEnumerator> ae;
  arr->Enumerate(getter_AddRefs(ae));
  PRUint32 l;
  arr->GetLength(&l);

  PRUint32 i;
  offset = 0.0;
  multiplier = 1.0;

  nsPrintfCString impCS("%u:", importNo);
  nsAutoString impS;
  CopyUTF8toUTF16(impCS, impS);

  for (i = 0; i < l; i++)
  {
    nsCOMPtr<nsISupports> unitSupports;
    ae->GetNext(getter_AddRefs(unitSupports));
    nsCOMPtr<nsIDOMElement> unitEl(do_QueryInterface(unitSupports));

    nsAutoString unitsName;
    unitEl->GetAttribute(NS_LITERAL_STRING("units"), unitsName);

    nsAutoString exponentS;
    nsresult rv = unitEl->GetAttribute(NS_LITERAL_STRING("exponent"), exponentS);
    double exponent = 1.0;
    if (rv == NS_OK && !exponentS.Equals(NS_LITERAL_STRING("")))
    {
      nsCAutoString exponentCS;
      CopyUTF16toUTF8(exponentS, exponentCS);
      exponent = PR_strtod(exponentCS.get(), nsnull);
      if (exponent == 0)
      {
        continue;
      }
    }

    nsCOMPtr<mozCellMLUnits> parentUnits;
    // Look for it in this component...
    allUnits.Get(impS + component + NS_LITERAL_STRING(":") + unitsName,
                 getter_AddRefs(parentUnits));
    if (parentUnits == nsnull)
    {
      // Try the model(including imports)...
      allUnits.Get(impS + NS_LITERAL_STRING(":") + unitsName,
                   getter_AddRefs(parentUnits));
      if (parentUnits == nsnull)
      {
        // Built-in?
        allUnits.Get(unitsName, getter_AddRefs(parentUnits));
        if (parentUnits == nsnull)
        {
          nsCAutoString ccomponent, cname;
          CopyUTF16toUTF8(component, ccomponent);
          CopyUTF16toUTF8(unitsName, cname);
          COMPLAIN(model, (T(model, "cantFindUnits"), cname.get(),
                           ccomponent.get()));
        }
      }
    }

    if (parentUnits->IsSeen() &&
        !parentUnits->IsDone())
    {
      // We have a cycle of units. Better complain...
      nsCAutoString name1, name2;
      CopyUTF16toUTF8(name, name1);
      CopyUTF16toUTF8(unitsName, name2);
      COMPLAIN(model, (T(model, "cyclicUnits"), name1.get(), name2.get()));
    }
    if (!parentUnits->IsDone())
    {
      nsresult rv = parentUnits->Initialise(model, importNumbers, allUnits);
      NS_ENSURE_SUCCESS(rv, rv);
    }

    if (l == 1)
    {
      nsAutoString offsetStr;
      nsresult rv = unitEl->GetAttribute(NS_LITERAL_STRING("offset"), offsetStr);
      if (rv == NS_OK)
      {
        nsCAutoString coffsetStr;
        CopyUTF16toUTF8(offsetStr, coffsetStr);
        offset += PR_strtod(coffsetStr.get(), nsnull);
      }
      offset += parentUnits->GetOffset();
    }

    double thisMup = 1;
    nsAutoString mupStr;
    rv = unitEl->GetAttribute(NS_LITERAL_STRING("multiplier"), mupStr);
    if (rv == NS_OK && !mupStr.Equals(NS_LITERAL_STRING("")))
    {
      nsCAutoString cmupStr;
      CopyUTF16toUTF8(mupStr, cmupStr);
      thisMup *= PR_strtod(cmupStr.get(), nsnull);
    }
    rv = unitEl->GetAttribute(NS_LITERAL_STRING("prefix"), mupStr);
    if (rv == NS_OK)
      thisMup *= PrefixToMultiplier(mupStr);

    thisMup *= parentUnits->GetMultiplier();

    multiplier *= pow(thisMup, exponent);

    nsCAutoString cunit, cname;
    CopyUTF16toUTF8(name, cname);
    CopyUTF16toUTF8(unitsName, cunit);

    // We want to keep it in the simplest form, so we must check if it already
    // exists before adding...
    PRUint32 j;
    if (parentUnits->GetIsBase())
    {
      for (j = 0; j < (PRUint32)unitList.Count(); j++)
        if (unitList[j]->base == parentUnits)
        {
          unitList[j]->exponent += exponent;
          break;
        }
      if (j == (PRUint32)unitList.Count())
      {
        nsCOMPtr<UnitBase> ub = new UnitBase(parentUnits, exponent);
        unitList.AppendObject(ub);
      }
    }
    else for (j = 0; j < parentUnits->GetUnitCount(); j++)
    {
      nsCOMPtr<UnitBase> baseUnit;
      baseUnit = parentUnits->GetUnit(j);
      PRUint32 k;
      for (k = 0; k < (PRUint32)unitList.Count(); k++)
        if (unitList[k]->base == baseUnit->base)
        {
          unitList[k]->exponent += exponent * baseUnit->exponent;
          break;
        }
      if (k == (PRUint32)unitList.Count())
      {
        nsCOMPtr<UnitBase> ub = new UnitBase(baseUnit->base,
                                             baseUnit->exponent * exponent);
        unitList.AppendObject(ub);
      }
    }
  }
  
  nsCAutoString cname;
  CopyUTF16toUTF8(name, cname);
#ifdef DEBUG_UNITS
  printf("Units list for %s is %u long.\n", cname.get(), unitList.Count());
#endif

  if (unitList.Count() >= 1)
    SortBases(0, unitList.Count() - 1);
  done = PR_TRUE;
  return NS_OK;
}

double
mozCellMLUnits::PrefixToMultiplier(nsAString& prefix)
{
#define FP(x, v) if (prefix.Equals(NS_LITERAL_STRING(#x))) {return v; }
#define P(x, v) else FP(x, v)
  // XXX I flipped all the exponent signs here so they correspond to the
  // multiplier according to the standard. However, the standard isn't clear on
  // this.
  if (prefix.Equals(NS_LITERAL_STRING("")))
    return 1E0;
  P(yotta, 1E-24)
  P(zetta, 1E-21)
  P(exa, 1E-18)
  P(peta, 1E-15)
  P(tera, 1E-12)
  P(giga, 1E-9)
  P(mega, 1E-6)
  P(kilo, 1E-3)
  P(hecto, 1E-2)
  P(deka, 1E-1)
  P(deci, 1E1)
  P(centi, 1E2)
  P(milli, 1E3)
  P(micro, 1E6)
  P(nano, 1E9)
  P(pico, 1E12)
  P(femto, 1E15)
  P(atto, 1E18)
  P(zepto, 1E21)
  P(yocto, 1E24)
#undef FP
#undef P
  nsCAutoString cprefix;
  CopyUTF16toUTF8(prefix, cprefix);
  double v = pow(10.0, -PR_strtod(cprefix.get(), nsnull));
  return v;
}

PRBool
mozCellMLUnits::Equals(mozCellMLUnits* cf)
{
  if (isbase)
  {
    if (cf->GetIsBase())
      return (cf == this);
    if (cf->GetOffset() != 0.0 ||
        cf->GetMultiplier() != 1.0)
      return PR_FALSE;
  }
  else if (cf->GetIsBase())
  {
    if (offset != 0.0 ||
        multiplier != 1.0)
      return PR_FALSE;
  }
  else if (cf->GetOffset() != offset)
    return PR_FALSE;
  else if (cf->GetMultiplier() != multiplier)
    return PR_FALSE;
  return Equivalent(cf);
}

PRBool
mozCellMLUnits::Equivalent(mozCellMLUnits* cf)
{
  PRUint32 i, l;
  if (isbase)
  {
    if (cf->GetIsBase())
      return (cf == this);
    if (cf->GetUnitCount() != 1)
      return PR_FALSE;
    UnitBase* ub = cf->GetUnit(0);
    if (ub->exponent != 1.0)
      return PR_FALSE;
    return (ub->base == this);
  }
  else if (cf->GetIsBase())
  {
    if (unitList.Count() != 1)
      return PR_FALSE;
    UnitBase* ub = unitList[0];
    if (ub->exponent != 1.0)
      return PR_FALSE;
    return (ub->base == cf);
  }
  l = unitList.Count();
#ifdef DEBUG_UNITS
  printf("Equivalent: Base Count = %u vs %u\n",
         l, cf->GetUnitCount());
#endif
  if (cf->GetUnitCount() != l)
    return PR_FALSE;
  for (i = 0; i < l; i++)
  {
#ifdef DEBUG_UNITS
    printf("Equivalent: Base Exponent = %e vs %e\n",
           cf->GetUnit(i)->exponent,
           unitList[i]->exponent);
#endif
    if (cf->GetUnit(i)->exponent != unitList[i]->exponent)
      return PR_FALSE;
    if (cf->GetUnit(i)->base != unitList[i]->base)
      return PR_FALSE;
  }
  return PR_TRUE;
}
